# Adapted from https://github.com/meshy/colour-runner by Charlie Denton
# License: MIT
#
from unittest.result import TestResult
from unittest.runner import TextTestRunner
from unittest.util import strclass

from blessings import Terminal
from pygments import formatters, highlight
from pygments.lexers import Python3TracebackLexer as Lexer


class ColourTextTestResult(TestResult):
    """
    A test result class that prints colour formatted text results to a stream.

    Based on https://github.com/python/cpython/blob/3.3/Lib/unittest/runner.py
    """
    formatter = formatters.Terminal256Formatter()
    lexer = Lexer()
    separator1 = '=' * 70
    separator2 = '-' * 70
    indent = ' ' * 4
    # if `checkmode` is true, simplified output will be generated with
    # no traceback
    checkmode = False
    _terminal = Terminal()
    colours = {
        None: str,
        'error': _terminal.bold_red,
        'expected': _terminal.blue,
        # 'fail': _terminal.bold_yellow,
        'fail': _terminal.bold_magenta,
        'skip': str,
        'success': _terminal.green,
        'title': _terminal.blue,
        'unexpected': _terminal.bold_red,
    }

    _test_class = None

    def __init__(self, stream, descriptions, verbosity):
        super(ColourTextTestResult, self).__init__(
            stream, descriptions, verbosity)
        self.stream = stream
        self.showAll = verbosity > 1
        self.dots = verbosity == 1
        self.descriptions = descriptions

    def getShortDescription(self, test):
        doc_first_line = test.shortDescription()
        if self.descriptions and doc_first_line:
            return self.indent + doc_first_line
        return self.indent + test._testMethodName

    def getLongDescription(self, test):
        doc_first_line = test.shortDescription()
        if self.descriptions and doc_first_line:
            return '\n'.join((str(test), doc_first_line))
        return str(test)

    def getClassDescription(self, test):
        test_class = test.__class__
        doc = test_class.__doc__
        if self.descriptions and doc:
            return doc.split('\n')[0].strip()
        return strclass(test_class)

    def startTest(self, test):
        super(ColourTextTestResult, self).startTest(test)
        pos = 0
        if self.showAll:
            if self._test_class != test.__class__:
                self._test_class = test.__class__
                title = self.getClassDescription(test)
                self.stream.writeln(self.colours['title'](title))
            descr = self.getShortDescription(test)
            self.stream.write(descr)
            pos += len(descr)
            self.stream.write(' ' * (70 - pos))
            # self.stream.write(' ' * (self._terminal.width - 10 - pos))
            # self.stream.write(' ... ')
            self.stream.flush()

    def printResult(self, short, extended, colour_key=None):
        colour = self.colours[colour_key]
        if self.showAll:
            self.stream.writeln(colour(extended))
        elif self.dots:
            self.stream.write(colour(short))
            self.stream.flush()

    def addSuccess(self, test):
        super(ColourTextTestResult, self).addSuccess(test)
        self.printResult('.', 'ok', 'success')

    def addError(self, test, err):
        super(ColourTextTestResult, self).addError(test, err)
        self.printResult('E', 'ERROR', 'error')

    def addFailure(self, test, err):
        super(ColourTextTestResult, self).addFailure(test, err)
        self.printResult('F', 'FAIL', 'fail')

    def addSkip(self, test, reason):
        super(ColourTextTestResult, self).addSkip(test, reason)
        if self.checkmode:
            self.printResult('s', 'skipped', 'skip')
        else:
            self.printResult('s', 'skipped {0!r}'.format(reason), 'skip')

    def addExpectedFailure(self, test, err):
        super(ColourTextTestResult, self).addExpectedFailure(test, err)
        self.printResult('x', 'expected failure', 'expected')

    def addUnexpectedSuccess(self, test):
        super(ColourTextTestResult, self).addUnexpectedSuccess(test)
        self.printResult('u', 'unexpected success', 'unexpected')

    def printErrors(self):
        if self.dots or self.showAll:
            self.stream.writeln()
        self.printErrorList('ERROR', self.errors)
        self.printErrorList('FAIL', self.failures)

    def printErrorList(self, flavour, errors):
        colour = self.colours[flavour.lower()]

        for test, err in errors:
            if self.checkmode and flavour == 'FAIL':
                self.stream.writeln(self.separator1)
                title = '%s: %s' % (flavour, test.shortDescription())
                self.stream.writeln(colour(title))
                self.stream.writeln(str(test))
                if self.showAll:
                    self.stream.writeln(self.separator2)
                    lines = str(err).split('\n')
                    i = 1
                    for line in lines[1:]:
                        if line.startswith(' '):
                            i += 1
                        else:
                            break
                    self.stream.writeln(highlight(
                        '\n'.join(lines[i:]), self.lexer, self.formatter))
            else:
                self.stream.writeln(self.separator1)
                title = '%s: %s' % (flavour, self.getLongDescription(test))
                self.stream.writeln(colour(title))
                self.stream.writeln(self.separator2)
                self.stream.writeln(highlight(err, self.lexer, self.formatter))


class ColourTextTestRunner(TextTestRunner):
    """A test runner that uses colour in its output"""
    resultclass = ColourTextTestResult



# -*- coding: utf-8 -*-
"""
A nested dict with both attribute and item access.

NA stands for Nested and Attribute.
"""
import collections
import copy


class NADict:
    """A nested dict with both attribute and item access.

    It is intended to be used with keys that are valid Python
    identifiers.  However, except for string keys containing a dot,
    there are actually no hard limitations.  If a key equals an existing
    attribute name, attribute access is of cause not possible.

    Nested items can be accessed via a dot notation, as shown in the
    example below.

    Examples
    --------
    >>> n = NADict(a=1, b=NADict(c=3, d=4))
    >>> n['a']
    1
    >>> n.a
    1
    >>> n['b.c']
    3
    >>> n.b.c
    3
    >>> n['b.e'] = 5
    >>> n.b.e
    5

    Attributes
    ----------
    _dict : dict
        Dictionary holding the actial items.
    """
    def __init__(self, *args, **kw):
        object.__setattr__(self, '_dict', {})
        self.update(*args, **kw)

    def __getitem__(self, key):
        if '.' in key:
            k1, k2 = key.split('.', 1)
            return self._dict[k1][k2]
        else:
            return self._dict[key]

    def __setitem__(self, key, value):
        if key in ('clear', 'copy', 'fromkeys', 'get', 'items', 'keys',
                   'pop', 'popitem', 'setdefault', 'update', 'values'):
            raise ValueError('invalid key "%s": must not override supported '
                             'dict method names' % key)
        elif '.' in key:
            k1, k2 = key.split('.', 1)
            if k1 not in self._dict:
                self._dict[k1] = NADict()
            self._dict[k1][k2] = value
        elif key in self._dict:
            if isinstance(self._dict[key], NADict):
                self._dict[key].update(value)
            else:
                self._dict[key] = value
        else:
            if isinstance(value, collections.abc.Mapping):
                self._dict[key] = NADict(value)
            else:
                self._dict[key] = value

    def __delitem__(self, key):
        if '.' in key:
            k1, k2 = key.split('.', 1)
            del self._dict[k1][k2]
        else:
            del self._dict[key]

    def __getattr__(self, key):
        if key not in self._dict:
            raise AttributeError('No such key: %s' % (key, ))
        return self._dict[key]

    def __setattr__(self, key, value):
        if key in self._dict:
            self._dict[key] = value
        else:
            object.__setattr__(self, key, value)

    def __delattr__(self, key):
        if key in self._dict:
            del self._dict[key]
        else:
            object.__delattr__(self, key)

    def __len__(self):
        return len(self._dict)

    def __contains__(self, key):
        if '.' in key:
            k1, k2 = key.split('.', 1)
            return k2 in self._dict[k1]
        else:
            return key in self._dict

    def __iter__(self, prefix=''):
        for k, v in self._dict.items():
            key = '%s.%s' % (prefix, k) if prefix else k
            if isinstance(v, NADict):
                yield from v.__iter__(key)
            else:
                yield key

    def __repr__(self):
        return '%s(%s)' % (
            self.__class__.__name__,
            ', '.join('%s=%s' % (k, repr(v)) for k, v in self._dict.items()))

    def clear(self):
        """Clear all keys."""
        self._dict.clear()

    def copy(self):
        """Returns a deep copy of self."""
        return copy.deepcopy(self)

    @staticmethod
    def fromkeys(self, iterable, value=None):
        """Returns a new NADict with keys from `iterable` and values
        set to `value`."""
        n = NADict()
        for key in iterable:
            n[key] = value
        return n

    def get(self, key, default=None):
        """Returns the value for `key` if `key` is in self, else return
        `default`."""
        if '.' in key:
            k1, k2 = key.split('.', 1)
            return self._dict[k1].get(k2, default)
        else:
            return self._dict.get(key, default)

    def items(self, prefix=''):
        """Returns an iterator over all items as (key, value) pairs."""
        for k, v in self._dict.items():
            key = '%s.%s' % (prefix, k) if prefix else k
            if isinstance(v, NADict):
                yield from v.items(key)
            else:
                yield (key, v)

    def keys(self, prefix=''):
        """Returns an iterator over all keys."""
        for k, v in self._dict.items():
            key = '%s.%s' % (prefix, k) if prefix else k
            if isinstance(v, NADict):
                yield from v.keys(key)
            else:
                yield key

    def pop(self, key, default=None):
        """Removed `key` and returns corresponding value.  If `key` is not
        found, `default` is returned if given, otherwise KeyError is
        raised."""
        if '.' in key:
            k1, k2 = key.split('.', 1)
            return self._dict[k1].pop(k2, default)
        else:
            return self._dict.pop(key, default)

    def popitem(self, prefix=''):
        """Removes and returns some (key, value). Raises KeyError if empty."""
        item = self._dict.popitem()
        if isinstance(item, NADict):
            k, v = item
            item2 = item.popitem(k)
            self._dict[k] = v
            return item2
        else:
            k, v = self._dict.popitem()
            key = '%s.%s' % (prefix, k) if prefix else k
            return (key, v)

    def setdefault(self, key, value=None):
        """Inserts `key` and `value` pair if key is not found.

        Returns the new value for `key`."""
        if '.' in key:
            k1, k2 = key.split('.', 1)
            return self._dict[k1].setdefault(k2, value)
        else:
            return self._dict.setdefault(key, value)

    def update(self, *args, **kw):
        """Updates self with dict/iterable from `args` and keyword arguments
        from `kw`."""
        for arg in args:
            if hasattr(arg, 'keys'):
                for k in arg:
                    self[k] = arg[k]
            else:
                for k, v in arg:
                    self[k] = v
        for k, v in kw.items():
            self[k] = v

    def values(self):
        """Returns a set-like providing a view of all style values."""
        return self._dict.values()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Step 4 - map to common representation
-------------------------------------
maps it to a collection of instances of the metadata generated from the
ontology.


This script uses the Atomistic Simulation Environment (ASE) to load
a atomistic Al-Fe4Al13 interface structure from a cif file and
represents it using the same metadata framework as used in step 1.
"""
import dlite


def map_app2common(inst, metacoll, out_id=None):
    """Maps atom structure `inst` from our application representation
    (based on a not explicitly stated ontology) to the common
    EMMO-based representation in `metacoll`.

    Parameters
    ----------
    inst : Instance of http://sintef.no/meta/soft/0.1/Atoms
        Input atom structure.
    metacoll : Collection
        Collection of EMMO-based metadata generated from the ontology.
    out_id : None | string
        An optional id associated with the returned collection.

    Returns
    -------
    atcoll : Collection
        New collection with the atom structure represented as instances
        of metadata in `metacoll`.

    Notes
    -----
    We use lowercase and underscore notation for the individuals.
    """
    infodict = dict(inst.info)  # make dict out of the info field

    # Create new collection representing `inst` in our case ontology
    atcoll = dlite.Collection(out_id)

    # Get metadata from metacoll
    Crystal = metacoll['Crystal']
    UnitCell = metacoll['CrystalUnitCell']
    EBondedAtom = metacoll['BondedAtom']

    # Instanciate the structure
    crystal = Crystal([])
    crystal.spacegroup = infodict['spacegroup']
    atcoll.add('crystal', crystal)

    unit_cell = UnitCell([3, 3, 36])
    unit_cell.lattice_vector = inst.cell
    atcoll.add('unit_cell', unit_cell)
    atcoll.add_relation('crystal', 'hasSpatialDirectPart', 'unit_cell')

    for i in range(inst.natoms):
        label = 'atom%d' % i
        a = EBondedAtom([3])
        a.AtomicNumber = inst.numbers[i]
        a.Position = inst.positions[i]
        atcoll.add(label, a)
        atcoll.add_relation('unit_cell', 'hasSpatialDirectPart', label)

    return atcoll


# Load metadata collection from step 1
metacoll = dlite.Collection(
    'json://usercase_metadata.json?mode=r#usercase_ontology', True)

# Load dlite-representation of atoms structure from step 3
coll = dlite.Collection(
    'json://usercase_appdata.json?mode=r#usercase_appdata', False)
inst = coll.get('atoms')

# Do the mapping
new = map_app2common(inst, metacoll)


# Append the new atoms collection to the storage
new.save('json://usercase_data.json?mode=w')



#!/usr/bin/env python3
import sys
import os

# Add emmo to sys path
thisdir = os.path.abspath(os.path.dirname(__file__))
sys.path.insert(1, os.path.abspath(os.path.join(thisdir, '..', '..')))
from emmo import get_ontology  # noqa: E402, F401
from emmo.ontodoc import OntoDoc, DocPP  # noqa: E402, F401


emmo = get_ontology()
emmo.load()

baseiri = 'http://emmo.info/'
iris = set(c.namespace.base_iri for c in emmo.classes())
iris.update(set(c.namespace.base_iri for c in emmo.object_properties()))
# iris.update(set(c.namespace.base_iri for c in emmo.annotation_properties()))

for s in sorted(iris):
    print(s)


inputdir = os.path.abspath(os.path.join(
    thisdir, '..', '..', 'examples', 'emmodoc'))
inputfile = os.path.join(thisdir, 'doc.md')

ontodoc = OntoDoc(emmo)

with open(inputfile, 'rt') as f:
    template = f.read()
docpp = DocPP(template, ontodoc, os.path.dirname(inputfile))
docpp.process()
with open('doc-output.md', 'wt') as f:
    f.write(docpp.get_buffer())



#!/usr/bin/env python3
import sys
import os

# Add emmo to sys path
thisdir = os.path.abspath(os.path.dirname(__file__))
sys.path.insert(1, os.path.abspath(os.path.join(thisdir, '..', '..')))
from emmo import get_ontology  # noqa: E402, F401

from emmo.graph import (plot_modules, get_module_dependencies,
                        check_module_dependencies)  # noqa: E402, F401


iri = 'http://emmo.info/emmo/1.0.0-alpha2'
emmo = get_ontology(iri)
emmo.load()

modules = get_module_dependencies(emmo)
plot_modules(modules, filename='modules.png')
check_module_dependencies(modules)



#!/usr/bin/env python3
import sys
import os

# Add emmo to sys path
thisdir = os.path.abspath(os.path.dirname(__file__))
sys.path.insert(1, os.path.abspath(os.path.join(thisdir, '..', '..')))
from emmo import get_ontology, World  # noqa: E402, F401

import owlready2  # noqa: E402, F401


emmodir = os.path.join(thisdir, '..', '..', '..', 'EMMO')
owlready2.set_log_level(0)


# Load latest stable EMMO
emmo = get_ontology('emmo')
emmo.load()


# Load emmo-inferred
world = World()
inferred = world.get_ontology()
inferred.load()


# Load local EMMO
if os.path.exists(os.path.join(emmodir, 'emmo.owl')):
    world2 = World()
    local = world2.get_ontology(os.path.join(emmodir, 'emmo.owl'))
    local.load(catalog_file=True)



# -*- coding: utf-8 -*-
import sys

__version__ = '1.0.0-alpha-22'

# Ensure correct Python version
if sys.version_info < (3, 6):
    raise RuntimeError('emmo requires Python 3.6 or later')

# Ensure emmo is imported before owlready2...
if 'owlready2' in sys.modules.keys():
    raise RuntimeError('emmo must be imported before owlready2')

# Monkey patch Owlready2 by injecting some methods
from . import patch  # noqa: E402, F401

# Import World and get_ontology(), which are our main entry points
from .ontology import World, get_ontology  # noqa: E402, F401

# Global list of ontology search paths
from owlready2 import onto_path  # noqa: E402, F401



# -*- coding: utf-8 -*-
"""
A module for visualising ontologies using graphviz.
"""
import os
import re
import tempfile
import xml.etree.ElementTree as ET

import owlready2
import graphviz

from .utils import asstring
from .ontology import get_ontology

typenames = owlready2.class_construct._restriction_type_2_label


# Literal for `root` arguments
ALL = 1


def getlabel(e):
    """Returns the label of entity `e`."""
    if hasattr(e, 'prefLabel'):
        return e.prefLabel.first()
    elif hasattr(e, '__name__'):
        return e.__name__
    elif hasattr(e, 'name'):
        return str(e.name)
    else:
        return repr(e)



class OntoGraph:
    """Class for visualising an ontology.

    Parameters
    ----------
    ontology : emmo.Ontology instance
        Ontology to visualize.
    root : None | graph.ALL | string | owlready2.ThingClass instance
        Name or owlready2 entity of root node to plot subgraph
        below.  If `root` is `graph.ALL`, all classes will be included
        in the subgraph.
    leafs : None | sequence
        A sequence of leaf node names for generating sub-graphs.
    entities : None | sequence
        A sequence of entities to add to the graph.
    relations : "all" | str | None | sequence
        Sequence of relations to visualise.  If "all", means to include
        all relations.
    style : None | dict | "default"
        A dict mapping the name of the different graphical elements
        to dicts of dot graph attributes. Supported graphical elements
        include:
          - graphtype : "Digraph" | "Graph"
          - graph : graph attributes (G)
          - class : nodes for classes (N)
          - root : additional attributes for root nodes (N)
          - leaf : additional attributes for leaf nodes (N)
          - defined_class : nodes for defined classes (N)
          - class_construct : nodes for class constructs (N)
          - individual : nodes for invididuals (N)
          - object_property : nodes for object properties (N)
          - data_property : nodes for data properties (N)
          - annotation_property : nodes for annotation properties (N)
          - added_node : nodes added because `addnodes` is true (N)
          - isA : edges for isA relations (E)
          - not : edges for not class constructs (E)
          - equivalent_to : edges for equivalent_to relations (E)
          - disjoint_with : edges for disjoint_with relations (E)
          - inverse_of : edges for inverse_of relations (E)
          - default_relation : default edges relations and restrictions (E)
          - relations : dict of styles for different relations (E)
          - inverse : default edges for inverse relations (E)
          - default_dataprop : default edges for data properties (E)
          - nodes : attribute for individual nodes (N)
          - edges : attribute for individual edges (E)
        If style is None or "default", the default style is used.
        See https://www.graphviz.org/doc/info/attrs.html
    edgelabels : bool | dict
        Whether to add labels to the edges of the generated graph.
        It is also possible to provide a dict mapping the
        full labels (with cardinality stripped off for restrictions)
        to some abbriviations.
    addnodes : bool
        Whether to add missing target nodes in relations.
    addconstructs : bool
        Whether to add nodes representing class constructs.
    parents : int
        Include `parents` levels of parents.
    excluded_nodes : None | sequence
        Sequence of labels of nodes to exclude.
    graph : None | pydot.Dot instance
        Graphviz Digraph object to plot into.  If None, a new graph object
        is created using the keyword arguments.
    imported : bool
        Whether to include imported classes if `entities` is None.
    kwargs :
        Passed to graphviz.Digraph.
    """
    _default_style = {
        'graphtype': 'Digraph',
        'graph': {
            'rankdir': 'BT', 'fontsize': '8',
            # 'fontname': 'Bitstream Vera Sans', 'splines': 'ortho',
        },
        'class': {
            'style': 'filled',
            'fillcolor': '#ffffcc',
        },
        'root': {'penwidth': '2'},
        'leaf': {'penwidth': '2'},
        'defined_class': {
            'style': 'filled',
            'fillcolor': '#ffc880',
        },
        'class_construct': {
            'style': 'filled',
            'fillcolor': 'gray',
        },
        'individual': {
            'shape': 'diamond',
            'style': 'filled',
            'fillcolor': '#874b82',
            'fontcolor': 'white',
        },
        'object_property': {
            'shape': 'box',
            'style': 'filled',
            'fillcolor': '#0079ba',
            'fontcolor': 'white',
        },
        'data_property': {
            'shape': 'box',
            'style': 'filled',
            'fillcolor': 'green',
        },
        'annotation_property': {
            'shape': 'box',
            'style': 'filled',
            'fillcolor': 'orange',
        },
        'added_node': {
            'color': 'red',
        },
        'isA': {'arrowhead': 'empty'},
        'not': {'color': 'gray', 'style': 'dotted'},
        'equivalent_to': {'color': 'green3'},
        'disjoint_with': {'color': 'red', 'constraint': 'false'},
        'inverse_of': {'color': 'orange', },
        'default_relation': {'color': 'olivedrab', 'constraint': 'false'},
        'relations': {
            'disconnected': {'color': 'red', 'style': 'dotted',
                             'arrowhead': 'odot'},
            'hasPart': {'color': 'blue'},
            'hasProperPart': {'color': 'blue', 'style': 'dashed'},
            'hasParticipant': {'color': 'red'},
            'hasProperParticipant': {'color': 'red', 'style': 'dashed'},
            'hasSpatialDirectPart': {'color': 'darkgreen'},
            'hasTemporalPart': {'color': 'magenta'},
            'hasTemporalDirectPart': {'color': 'magenta', 'style': 'dashed'},
            'hasReferenceUnit': {'color': 'darkgreen', 'style': 'dashed'},
            'hasSign': {'color': 'orange'},
            'hasConvention': {'color': 'orange', 'style': 'dashed'},
            'hasProperty': {'color': 'orange', 'style': 'dotted'},
        },
        'inverse': {'arrowhead': 'inv'},
        'default_dataprop': {'color': 'green', 'constraint': 'false'},
        'node': {},
        'edge': {},
    }

    def __init__(self, ontology, root=None, leafs=None, entities=None,
                 relations='isA', style=None, edgelabels=True,
                 addnodes=False, addconstructs=False,
                 parents=0, excluded_nodes=None, graph=None,
                 imported=False, **kwargs):
        if style is None or style == 'default':
            style = self._default_style

        if graph is None:
            graphtype = style.get('graphtype', 'Digraph')
            dotcls = getattr(graphviz, graphtype)
            graph_attr = kwargs.pop('graph_attr', {})
            for k, v in style.get('graph', {}).items():
                graph_attr.setdefault(k, v)
            self.dot = dotcls(graph_attr=graph_attr, **kwargs)
            self.nodes = set()
            self.edges = set()
        else:
            if ontology != graph.ontology:
                ValueError(
                    'the same ontology must be used when extending a graph')
            self.dot = graph.dot.copy()
            self.nodes = graph.nodes.copy()
            self.edges = graph.edges.copy()

        self.ontology = ontology
        self.relations = set(
            [relations] if isinstance(relations, str) else relations)
        self.style = style
        self.edgelabels = edgelabels
        self.addnodes = addnodes
        self.addconstructs = addconstructs
        self.excluded_nodes = set(excluded_nodes) if excluded_nodes else set()
        self.imported = imported

        if root == ALL:
            self.add_entities(
                relations=relations, edgelabels=edgelabels,
                addnodes=addnodes, addconstructs=addconstructs)
        elif root:
            self.add_branch(
                root, leafs,
                relations=relations, edgelabels=edgelabels,
                addnodes=addnodes, addconstructs=addconstructs)
            if parents:
                self.add_parents(
                    root, levels=parents,
                    relations=relations, edgelabels=edgelabels,
                    addnodes=addnodes, addconstructs=addconstructs)

        if entities:
            self.add_entities(entities=entities, relations=relations,
                              edgelabels=edgelabels, addnodes=addnodes,
                              addconstructs=addconstructs)

    def add_entities(self, entities=None, relations='isA', edgelabels=True,
                     addnodes=False, addconstructs=False,
                     nodeattrs=None, **attrs):
        """Adds a sequence of entities to the graph.  If `entities` is None,
        all classes are added to the graph.

        `nodeattrs` is a dict mapping node names to are attributes for
        dedicated nodes.
        """
        if entities is None:
            entities = self.ontology.classes(imported=self.imported)
        self.add_nodes(entities, nodeattrs=nodeattrs, **attrs)
        self.add_edges(
            relations=relations, edgelabels=edgelabels,
            addnodes=addnodes, addconstructs=addconstructs, **attrs)

    def add_branch(self, root, leafs=None, include_leafs=True,
                   strict_leafs=False, exclude=None, relations='isA',
                   edgelabels=True, addnodes=False, addconstructs=False,
                   **attrs):
        """Adds branch under `root` ending at any entiry included in the
        sequence `leafs`.  If `include_leafs` is true, leafs classes are
        also included."""
        if leafs is None:
            leafs = ()

        classes = self.ontology.get_branch(
            root=root, leafs=leafs, include_leafs=include_leafs,
            strict_leafs=strict_leafs, exclude=exclude)

        nodeattrs = {}
        nodeattrs[getlabel(root)] = self.style.get('root', {})
        for leaf in leafs:
            nodeattrs[getlabel(leaf)] = self.style.get('leaf', {})

        self.add_entities(
            entities=classes,
            relations=relations, edgelabels=edgelabels,
            addnodes=addnodes, addconstructs=addconstructs,
            nodeattrs=nodeattrs, **attrs)

    def add_parents(self, name, levels=1, relations='isA',
                    edgelabels=None, addnodes=False, addconstructs=False,
                    **attrs):
        """Add `levels` levels of strict parents of entity `name`."""
        def addparents(e, n, s):
            if n > 0:
                for p in e.get_parents(strict=True):
                    s.add(p)
                    addparents(p, n - 1, s)
        e = self.ontology[name] if isinstance(name, str) else name
        parents = set()
        addparents(e, levels, parents)
        self.add_entities(
            entities=parents,
            relations=relations, edgelabels=edgelabels,
            addnodes=addnodes, addconstructs=addconstructs, **attrs)

    def add_node(self, name, nodeattrs=None, **attrs):
        """Add node with given name. `attrs` are graphviz node attributes."""
        e = self.ontology[name] if isinstance(name, str) else name
        label = getlabel(e)
        if label not in self.nodes.union(self.excluded_nodes):
            kw = self.get_node_attrs(e, nodeattrs=nodeattrs, attrs=attrs)
            if hasattr(e, 'iri'):
                kw.setdefault('URL', e.iri)
            self.dot.node(label, label=label, **kw)
            self.nodes.add(label)

    def add_nodes(self, names, nodeattrs, **attrs):
        """Add nodes with given names. `attrs` are graphviz node attributes."""
        for name in names:
            self.add_node(name, nodeattrs=nodeattrs, **attrs)

    def add_edge(self, subject, predicate, object, edgelabel=None, **attrs):
        """Add edge corresponding for ``(subject, predicate, object)``
        triplet."""
        subject = subject if isinstance(subject, str) else getlabel(subject)
        predicate = predicate if isinstance(predicate, str) else getlabel(
            predicate)
        object = object if isinstance(object, str) else getlabel(object)
        if subject in self.excluded_nodes or object in self.excluded_nodes:
            return
        if not isinstance(subject, str) or not isinstance(object, str):
            raise TypeError('`subject` and `object` must be strings')
        if subject not in self.nodes:
            raise RuntimeError('`subject` "%s" must have been added' % subject)
        if object not in self.nodes:
            raise RuntimeError('`object` "%s" must have been added' % object)
        key = (subject, predicate, object)
        if key not in self.edges:
            if edgelabel is None:
                edgelabel = self.edgelabels

            if isinstance(edgelabel, str):
                label = edgelabel
            if isinstance(edgelabel, dict):
                label = edgelabel.get(predicate, predicate)
            elif edgelabel:
                label = predicate
            else:
                label = None

            kw = self.get_edge_attrs(predicate, attrs=attrs)
            self.dot.edge(subject, object, label=label, **kw)
            self.edges.add(key)

    def add_source_edges(self, source, relations=None, edgelabels=None,
                         addnodes=None, addconstructs=None, **attrs):
        """Adds all relations originating from entity `source` who's type
        are listed in `relations`."""
        if relations is None:
            relations = self.relations
        elif isinstance(relations, str):
            relations = set([relations])
        else:
            relations = set(relations)

        edgelabels = self.edgelabels if edgelabels is None else edgelabels
        addconstructs = (
            self.addconstructs if addconstructs is None else addconstructs)

        e = self.ontology[source] if isinstance(source, str) else source
        label = getlabel(e)
        for r in e.is_a:

            # isA
            if isinstance(r, (owlready2.ThingClass,
                              owlready2.ObjectPropertyClass)):
                if 'all' in relations or 'isA' in relations:
                    rlabel = getlabel(r)
                    if r not in e.get_parents(strict=True):
                        continue
                    if not self.add_missing_node(r, addnodes=addnodes):
                        continue
                    self.add_edge(
                        subject=label, predicate='isA', object=rlabel,
                        edgelabel=edgelabels, **attrs)

            # restriction
            elif isinstance(r, owlready2.Restriction):
                rname = getlabel(r.property)
                if 'all' in relations or rname in relations:
                    rlabel = '%s %s' % (rname, typenames[r.type])
                    if isinstance(r.value, owlready2.ThingClass):
                        obj = getlabel(r.value)
                        if not self.add_missing_node(r.value, addnodes):
                            continue
                    elif (isinstance(r.value, owlready2.ClassConstruct) and
                          self.addconstructs):
                        obj = self.add_class_construct(r.value)
                    else:
                        continue
                    pred = asstring(r, exclude_object=True)
                    self.add_edge(label, pred, obj, edgelabel=edgelabels,
                                  **attrs)

            # inverse
            if isinstance(r, owlready2.Inverse):
                if 'all' in relations or 'inverse' in relations:
                    rlabel = getlabel(r)
                    if not self.add_missing_node(r, addnodes=addnodes):
                        continue
                    if r not in e.get_parents(strict=True):
                        continue
                    self.add_edge(
                        subject=label, predicate='inverse', object=rlabel,
                        edgelabel=edgelabels, **attrs)

    def add_edges(self, sources=None, relations=None, edgelabels=None,
                  addnodes=None, addconstructs=None, **attrs):
        """Adds all relations originating from entities `sources` who's type
        are listed in `relations`.  If `sources` is None, edges are added
        between all current nodes."""
        if sources is None:
            sources = self.nodes
        for source in sources.copy():
            self.add_source_edges(
                source, relations=relations, edgelabels=edgelabels,
                addnodes=addnodes, addconstructs=addconstructs, **attrs)

    def add_missing_node(self, name, addnodes=None):
        """Checks if `name` corresponds to a missing node and add it if
        `addnodes` is true.

        Returns true if the node exists or is added, false otherwise."""
        addnodes = self.addnodes if addnodes is None else addnodes
        e = self.ontology[name] if isinstance(name, str) else name
        label = getlabel(e)
        if label not in self.nodes:
            if addnodes:
                self.add_node(e, **self.style.get('added_node', {}))
            else:
                return False
        return True

    def add_class_construct(self, c):
        """Adds class construct `c` and return its label."""
        self.add_node(c, **self.style.get('class_construct', {}))
        label = getlabel(c)
        if isinstance(c, owlready2.Or):
            for cls in c.Classes:
                clslabel = getlabel(cls)
                if clslabel not in self.nodes and self.addnodes:
                    self.add_node(cls)
                if clslabel in self.nodes:
                    self.add_edge(getlabel(cls), 'isA', label)
        elif isinstance(c, owlready2.And):
            for cls in c.Classes:
                clslabel = getlabel(cls)
                if clslabel not in self.nodes and self.addnodes:
                    self.add_node(cls)
                if clslabel in self.nodes:
                    self.add_edge(label, 'isA', getlabel(cls))
        elif isinstance(c, owlready2.Not):
            clslabel = getlabel(c.Class)
            if clslabel not in self.nodes and self.addnodes:
                self.add_node(c.Class)
            if clslabel in self.nodes:
                self.add_edge(clslabel, 'not', label)
        # Neither and nor inverse constructs are
        return label

    def get_node_attrs(self, name, nodeattrs, attrs):
        """Returns attributes for node or edge `name`.  `attrs` overrides
        the default style."""
        e = self.ontology[name] if isinstance(name, str) else name
        label = getlabel(e)
        # class
        if isinstance(e, owlready2.ThingClass):
            if self.ontology.is_defined(e):
                kw = self.style.get('defined_class', {})
            else:
                kw = self.style.get('class', {})
        # class construct
        elif isinstance(e, owlready2.ClassConstruct):
            kw = self.style.get('class_construct', {})
        # individual
        elif isinstance(e, owlready2.Thing):
            kw = self.style.get('individual', {})
        # object property
        elif isinstance(e, owlready2.ObjectPropertyClass):
            kw = self.style.get('object_property', {})
        # data property
        elif isinstance(e, owlready2.DataPropertyClass):
            kw = self.style.get('data_property', {})
        # annotation property
        elif isinstance(e, owlready2.AnnotationPropertyClass):
            kw = self.style.get('annotation_property', {})
        else:
            raise TypeError('Unknown entity type: %r' % e)
        kw = kw.copy()
        kw.update(self.style.get('nodes', {}).get(label, {}))
        if nodeattrs:
            kw.update(nodeattrs.get(label, {}))
        kw.update(attrs)
        return kw

    def get_edge_attrs(self, predicate, attrs):
        """Returns attributes for node or edge `name`.  `attrs` overrides
        the default style."""
        # given type
        types = ('isA', 'equivalent_to', 'disjoint_with', 'inverse_of')
        if predicate in types:
            kw = self.style.get(predicate, {}).copy()
        else:
            kw = {}
            name = predicate.split(None, 1)[0]
            m = re.match(r'Inverse\((.*)\)', name)
            if m:
                name, = m.groups()
                attrs = attrs.copy()
                for k, v in self.style.get('inverse', {}).items():
                    attrs.setdefault(k, v)
            if not isinstance(name, str) or name in self.ontology:
                e = self.ontology[name] if isinstance(name, str) else name
                relations = self.style.get('relations', {})
                rels = set(self.ontology[r] for r in relations.keys()
                           if r in self.ontology)
                for r in e.mro():
                    if r in rels:
                        break
                rattrs = relations[getlabel(r)] if r in rels else {}
                # object property
                if isinstance(e, (owlready2.ObjectPropertyClass,
                                  owlready2.ObjectProperty)):
                    kw = self.style.get('default_relation', {}).copy()
                    kw.update(rattrs)
                # data property
                elif isinstance(e, (owlready2.DataPropertyClass,
                                    owlready2.DataProperty)):
                    kw = self.style.get('default_dataprop', {}).copy()
                    kw.update(rattrs)
                else:
                    raise TypeError('Unknown entity type: %r' % e)
        kw.update(self.style.get('edges', {}).get(predicate, {}))
        kw.update(attrs)
        return kw

    def add_legend(self, relations=None):
        """Adds legend for specified relations to the graph.

        If `relations` is "all", the legend will contain all relations
        that are defined in the style.  By default the legend will
        only contain relations that are currently included in the
        graph.

        Hence, you usually want to call add_legend() as the last method
        before saving or displaying.
        """
        rels = self.style.get('relations', {})
        if relations is None:
            relations = self.get_relations(sort=True)
        elif relations == 'all':
            relations = ['isA'] + list(rels.keys()) + ['inverse']
        elif isinstance(relations, str):
            relations = relations.split(',')

        n = len(relations)
        if n == 0:
            return

        t = ('<<table border="0" cellpadding="2" cellspacing="0" '
             'cellborder="0">')
        label1 = [t]
        label2 = [t]
        for i, r in enumerate(relations):
            label1.append(
                '<tr><td align="right" port="i%d">%s</td></tr>' % (i, r))
            label2.append('<tr><td port="i%d">&nbsp;</td></tr>' % i)
        label1.append('</table>>')
        label2.append('</table>>')
        self.dot.node('key1', label='\n'.join(label1), shape='plaintext')
        self.dot.node('key2', label='\n'.join(label2), shape='plaintext')

        rankdir = self.dot.graph_attr.get('rankdir', 'TB')
        constraint = 'false' if rankdir in ('TB', 'BT') else 'true'
        inv = True if rankdir in ('BT', ) else False

        for i in range(n):
            r = relations[n - 1 - i] if inv else relations[i]
            if r == 'inverse':
                kw = self.style.get('inverse', {}).copy()
            else:
                kw = self.get_edge_attrs(r, {}).copy()
            kw['constraint'] = constraint
            with self.dot.subgraph(name='sub%d' % i) as s:
                s.attr(rank='same')
                if rankdir in ('BT', 'LR'):
                    self.dot.edge('key1:i%d:e' % i, 'key2:i%d:w' % i, **kw)
                else:
                    self.dot.edge('key2:i%d:w' % i, 'key1:i%d:e' % i, **kw)

    def get_relations(self, sort=True):
        """Returns a set of relations in current graph.  If `sort` is true,
        a sorted list is returned."""
        relations = set()
        for s, p, o in self.edges:
            if p.startswith('Inverse'):
                relations.add('inverse')
                m = re.match(r'Inverse\((.+)\)', p)
                assert m
                relations.add(m.groups()[0])
            else:
                relations.add(p.split(None, 1)[0])

        # Sort, but place 'isA' first and 'inverse' last
        if sort:
            start, end = [], []
            if 'isA' in relations:
                relations.remove('isA')
                start.append('isA')
            if 'inverse' in relations:
                relations.remove('inverse')
                end.append('inverse')
            relations = start + sorted(relations) + end

        return relations

    def save(self, filename, format=None, **kwargs):
        """Saves graph to `filename`.  If format is not given, it is
        inferred from `filename`."""
        base, ext = os.path.splitext(filename)
        if format is None:
            format = ext.lstrip('.')
        kwargs.setdefault('cleanup', True)
        if format in ('graphviz', 'gv'):
            if 'dictionary' in kwargs:
                self.dot.save(filename, dictionary=kwargs['dictionary'])
            else:
                self.dot.save(filename)
        else:
            self.dot.render(base, format=format, **kwargs)

    def view(self):
        """Shows the graph in a viewer."""
        self.dot.view(cleanup=True)

    def get_figsize(self):
        """Returns the default figure size (width, height) in points."""
        with tempfile.TemporaryDirectory() as tmpdir:
            tmpfile = os.path.join(tmpdir, 'graph.svg')
            self.save(tmpfile)
            xml = ET.parse(tmpfile)
            svg = xml.getroot()
            width = svg.attrib['width']
            height = svg.attrib['height']
            assert width.endswith('pt')  # ensure that units are in points

            def asfloat(s):
                return float(re.match(r'^[\d.]+', s).group())
        return asfloat(width), asfloat(height)


def get_module_dependencies(iri_or_onto, strip_base=None):
    """Reads `iri_or_onto` and returns a dict mapping ontology names to a
    list of ontologies that they depends on.

    If `strip_base` is true, the base IRI is stripped from ontology
    names.  If it is a string, it lstrip'ped from the base iri.
    """
    if isinstance(iri_or_onto, str):
        onto = get_ontology(iri_or_onto)
        onto.load()
    else:
        onto = iri_or_onto

    modules = {onto.base_iri: set()}

    def strip(base_iri):
        if isinstance(strip_base, str):
            return base_iri.lstrip(strip_base)
        elif strip_base:
            return base_iri.strip(onto.base_iri)
        else:
            return base_iri

    def setmodules(onto):
        for o in onto.imported_ontologies:
            if onto.base_iri in modules:
                modules[strip(onto.base_iri)].add(strip(o.base_iri))
            else:
                modules[strip(onto.base_iri)] = set([strip(o.base_iri)])
            if o.base_iri not in modules:
                modules[strip(o.base_iri)] = set()
            setmodules(o)

    setmodules(onto)
    return modules


def plot_modules(src, filename=None, format=None, show=False,
                 strip_base=None, ignore_redundant=True):
    """Plot module dependency graph for `src` and return a graph object.

    Here `src` may be an IRI, a path the the ontology or a dict returned by
    get_module_dependencies().

    If `filename` is given, write the graph to this file.

    If `format` is None, the output format is inferred from `filename`.

    If `show` is true, the graph is displayed.

    `strip_base` is passed on to get_module_dependencies() if `src` is not
    a dict.

    If `ignore_redundant` is true, redundant dependencies are not plotted.
    """
    if isinstance(src, dict):
        modules = src
    else:
        modules = get_module_dependencies(src, strip_base=strip_base)

    if ignore_redundant:
        modules = check_module_dependencies(modules, verbose=False)

    dot = graphviz.Digraph(comment='Module dependencies')
    dot.attr(rankdir='TB')
    dot.node_attr.update(style='filled', fillcolor='lightblue', shape='box',
                         edgecolor='blue')
    dot.edge_attr.update(arrowtail='open', dir='back')

    for iri in modules.keys():
        iriname = iri.split(':', 1)[1]
        dot.node(iriname, label=iri, URL=iri)

    for iri, deps in modules.items():
        for dep in deps:
            iriname = iri.split(':', 1)[1]
            depname = dep.split(':', 1)[1]
            dot.edge(depname, iriname)

    if filename:
        base, ext = os.path.splitext(filename)
        if format is None:
            format = ext.lstrip('.')
        dot.render(base, format=format, view=False, cleanup=True)

    if show:
        dot.view(cleanup=True)

    return dot


def check_module_dependencies(modules, verbose=True):
    """Check module dependencies and return a copy of modules with
    redundant dependencies removed.

    If `verbose` is true, warnings are printed for each module that

    If `modules` is given, it should be a dict returned by
    get_module_dependencies().
    """
    def get_deps(iri, excl=None):
        """Returns a set with all dependencies of `iri`, excluding `excl` and
        its dependencies."""
        deps = set()
        for d in modules[iri]:
            if d != excl:
                deps.add(d)
                deps.update(get_deps(d))
        return deps

    mods = {}
    redundant = []
    for iri, deps in modules.items():
        if not deps:
            mods[iri] = set()
        for dep in deps:
            if dep in get_deps(iri, dep):
                redundant.append((iri, dep))
            elif iri in mods:
                mods[iri].add(dep)
            else:
                mods[iri] = set([dep])

    if redundant and verbose:
        print('** Warning: Redundant module dependency:')
        for iri, dep in redundant:
            print('%s -> %s' % (iri, dep))

    return mods



# -*- coding: utf-8 -*-
"""
A module for testing an ontology against conventions defined for EMMO.

A YAML file can be provided with additional test configurations.

Example configuration file:

    test_unit_dimensions:
      exceptions:
        - myunits.MyUnitCategory1
        - myunits.MyUnitCategory2
"""
import os
import sys
import re
import unittest
import itertools
import argparse
import fnmatch

from .ontology import World
from .patch import get_preferred_label as get_label
from . import onto_path

try:
    from .colortest import ColourTextTestRunner as TextTestRunner
except ImportError:
    from unittest import TextTestRunner


class TestEMMOConventions(unittest.TestCase):
    """Base class for testing an ontology against EMMO conventions."""
    config = {}  # configurations

    def get_config(self, string, default=None):
        """Returns the configuration specified by `string`.

        If configuration is not found in the configuration file, `default`
        is returned.

        Sub-configurations can be accessed by separating the components with
        dots, like "test_namespace.exceptions".
        """
        c = self.config
        try:
            for token in string.split('.'):
                c = c[token]
        except KeyError:
            return default
        return c


class TestSyntacticEMMOConventions(TestEMMOConventions):
    """Test syntactic EMMO conventions."""
    def test_number_of_labels(self):
        """Check that all entities have one and only one prefLabel.

        Use "altLabel" for synonyms.

        The only allowed exception is entities who's representation
        starts with "owl.".
        """
        exceptions = set((
            'terms.license',
            'terms.abstract',
            'terms.contributor',
            'terms.creator',
            'terms.publisher',
            'terms.title',
            'core.prefLabel',
            'core.altLabel',
            'core.hiddenLabel',
        ))
        exceptions.update(self.get_config('test_number_of_labels', ()))

        for e in self.onto.get_entities():
            if repr(e) not in exceptions:
                with self.subTest(entity=e, label=get_label(e),
                                  prefLabels=e.prefLabel):
                    if not repr(e).startswith('owl.'):
                        self.assertTrue(hasattr(e, 'prefLabel'))
                        self.assertEqual(1, len(e.prefLabel))

    def test_class_label(self):
        """Check that class labels are CamelCase and valid (Python) identifiers.

        For CamelCase, we are currently only checking that the labels
        start with upper case.
        """
        exceptions = set((
            '0-manifold',
            '1-manifold',
            '2-manifold',
            '3-manifold',
        ))
        for cls in self.onto.classes(self.check_imported):
            for label in cls.label + getattr(cls, 'prefLabel', []):
                if label not in exceptions:
                    with self.subTest(entity=cls, label=label):
                        self.assertTrue(label.isidentifier())
                        self.assertTrue(label[0].isupper())

    def test_object_property_label(self):
        """Check that object property labels are lowerCamelCase.

        Allowed exceptions: "EMMORelation"

        If they start with "has" or "is" they should be followed by a
        upper case letter.

        If they start with "is" they should also end with "Of".
        """
        for op in self.onto.object_properties():
            for label in op.label:
                if label == 'EMMORelation':
                    continue
                self.assertTrue(label[0].islower())
                if label.startswith('has'):
                    self.assertTrue(label[3].isupper())
                if label.startswith('is'):
                    self.assertTrue(label[2].isupper())
                    self.assertTrue(label.endswith('Of'))


class TestFunctionalEMMOConventions(TestEMMOConventions):
    """Test functional EMMO conventions."""

    def test_unit_dimension(self):
        """Check that all measurement units have a physical dimension.

        Configurations:
            exceptions - full class names of classes to ignore.
        """
        exceptions = set((
            'metrology.MultipleUnit',
            'metrology.SubMultipleUnit',
            'metrology.OffSystemUnit',
            'metrology.PrefixedUnit',
            'metrology.NonPrefixedUnit',
            'metrology.SpecialUnit',
            'metrology.DerivedUnit',
            'metrology.BaseUnit',
            'metrology.UnitSymbol',
            'siunits.SICoherentDerivedUnit',
            'siunits.SINonCoherentDerivedUnit',
            'siunits.SISpecialUnit',
            'siunits.SICoherentUnit',
            'siunits.SIPrefixedUnit',
            'siunits.SIBaseUnit',
            'siunits.SIUnitSymbol',
            'siunits.SIUnit',

            'emmo.MultipleUnit',
            'emmo.SubMultipleUnit',
            'emmo.OffSystemUnit',
            'emmo.PrefixedUnit',
            'emmo.NonPrefixedUnit',
            'emmo.SpecialUnit',
            'emmo.DerivedUnit',
            'emmo.BaseUnit',
            'emmo.UnitSymbol',

            'emmo.SICoherentDerivedUnit',
            'emmo.SINonCoherentDerivedUnit',
            'emmo.SISpecialUnit',
            'emmo.SICoherentUnit',
            'emmo.SIPrefixedUnit',
            'emmo.SIBaseUnit',
            'emmo.SIUnitSymbol',
            'emmo.SIUnit',
        ))
        exceptions.update(
            self.get_config('test_unit_dimension.exceptions', ()))
        regex = re.compile(r'^(emmo|metrology).hasPhysicalDimension.some\(.*\)$')
        classes = set(self.onto.classes(self.check_imported))
        for cls in self.onto.MeasurementUnit.descendants():
            if not self.check_imported and cls not in classes:
                continue
            # Assume that actual units are not subclassed
            if not list(cls.subclasses()) and repr(cls) not in exceptions:
                with self.subTest(cls=cls, label=get_label(cls)):
                    self.assertTrue(
                        any(regex.match(repr(r))
                            for r in cls.get_indirect_is_a()), msg=cls)

    def test_quantity_dimension(self):
        """Check that all quantities have a physicalDimension annotation.

        Configurations:
            exceptions - full class names of classes to ignore.
        """
        exceptions = set((
            'properties.ModelledQuantitativeProperty',
            'properties.MeasuredQuantitativeProperty',
            'properties.ConventionalQuantitativeProperty',
            'metrology.QuantitativeProperty',
            'metrology.Quantity',
            'metrology.OrdinalQuantity',
            'metrology.BaseQuantity',
            'metrology.PhysicalConstant',
            'metrology.PhysicalQuantity',
            'metrology.ExactConstant',
            'metrology.MeasuredConstant',
            'metrology.DerivedQuantity',
            'isq.ISQBaseQuantity',
            'isq.InternationalSystemOfQuantity',
            'isq.ISQDerivedQuantity',
            'isq.SIExactConstant',

            'emmo.ModelledQuantitativeProperty',
            'emmo.MeasuredQuantitativeProperty',
            'emmo.ConventionalQuantitativeProperty',

            'emmo.QuantitativeProperty',
            'emmo.Quantity',
            'emmo.OrdinalQuantity',
            'emmo.BaseQuantity',
            'emmo.PhysicalConstant',
            'emmo.PhysicalQuantity',
            'emmo.ExactConstant',
            'emmo.MeasuredConstant',
            'emmo.DerivedQuantity',

            'emmo.ISQBaseQuantity',
            'emmo.InternationalSystemOfQuantity',
            'emmo.ISQDerivedQuantity',
            'emmo.SIExactConstant',
        ))
        exceptions.update(
            self.get_config('test_quantity_dimension.exceptions', ()))
        regex = re.compile(
            '^T([+-][1-9]|0) L([+-][1-9]|0) M([+-][1-9]|0) I([+-][1-9]|0) '
            '(H|Θ)([+-][1-9]|0) N([+-][1-9]|0) J([+-][1-9]|0)$')
        classes = set(self.onto.classes(self.check_imported))
        for cls in self.onto.PhysicalQuantity.descendants():
            if not self.check_imported and cls not in classes:
                continue
            if repr(cls) not in exceptions:
                with self.subTest(cls=cls, label=get_label(cls)):
                    anno = cls.get_annotations()
                    self.assertIn('physicalDimension', anno, msg=cls)
                    physdim = anno['physicalDimension'].first()
                    self.assertRegex(physdim, regex, msg=cls)

    def test_namespace(self):
        """Check that all IRIs are namespaced after their (sub)ontology.

        Configurations:
            exceptions - full name of entities to ignore.
        """
        exceptions = set((
            'owl.qualifiedCardinality',
            'owl.minQualifiedCardinality',
            'terms.creator',
            'terms.contributor',
            'terms.publisher',
            'terms.title',
            'terms.license',
            'terms.abstract',
            'core.prefLabel',
            'core.altLabel',
            'core.hiddenLabel',
            'mereotopology.Item',
            'manufacturing.EngineeredMaterial',
        ))
        exceptions.update(self.get_config('test_namespace.exceptions', ()))
        def checker(onto, ignore_namespace):
            if list(filter(onto.base_iri.strip('#').endswith,
                           self.ignore_namespace)) != []:
                print('Skipping namespace: ' + onto.base_iri)
                return
            entities = itertools.chain(onto.classes(),
                                       onto.object_properties(),
                                       onto.data_properties(),
                                       onto.individuals(),
                                       onto.annotation_properties())
            for e in entities:
                if e not in visited and repr(e) not in exceptions:
                    visited.add(e)
                    with self.subTest(
                            iri=e.iri, base_iri=onto.base_iri, entity=repr(e)):
                        self.assertTrue(
                            e.iri.endswith(e.name),
                            msg='the final part of entity IRIs must be their '
                            'name')
                        self.assertEqual(
                            e.iri, onto.base_iri + e.name,
                            msg='IRI %r does not correspond to module '
                            'namespace: %r' % (e.iri, onto.base_iri))

            if self.check_imported:
                for imp_onto in onto.imported_ontologies:
                    if imp_onto not in visited_onto:
                        visited_onto.add(imp_onto)
                        checker(imp_onto, ignore_namespace)

        visited = set()
        visited_onto = set()
        checker(self.onto, self.ignore_namespace)


def main():
    """Run all checks on ontology `iri`.  Default is 'http://emmo.info/emmo'.
    """
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        'iri',
        help='File name or URI to the ontology to test.')
    parser.add_argument(
        '--database', '-d', metavar='FILENAME', default=':memory:',
        help='Load ontology from Owlready2 sqlite3 database.  The `iri` '
        'argument should in this case be the IRI of the ontology you '
        'want to check.')
    parser.add_argument(
        '--local', '-l', action='store_true',
        help='Load imported ontologies locally.  Their paths are specified '
        'in Protègè catalog files or via the --path option.  The IRI should '
        'be a file name.')
    parser.add_argument(
        '--catalog-file', default='catalog-v001.xml',
        help='Name of Protègè catalog file in the same folder as the '
        'ontology.  This option is used together with --local and '
        'defaults to "catalog-v001.xml".')
    parser.add_argument(
        '--path', action='append', default=[],
        help='Paths where imported ontologies can be found.  May be provided '
        'as a comma-separated string and/or with multiple --path options.')
    parser.add_argument(
        '--check-imported', '-i', action='store_true',
        help='Whether to check imported ontologies.')
    parser.add_argument(
        '--verbose', '-v', action='store_true',
        help='Verbosity level.')
    parser.add_argument(
        '--configfile', '-c',
        help='A yaml file with additional test configurations.')
    parser.add_argument(
        '--skip', '-s', action='append', default=[],
        help=('Shell pattern matching tests to skip.  This option may be '
              'provided multiple times.'))
    parser.add_argument(
        '--url-from-catalog', '-u', action='store_true',
        help=('Get url from catalog file'))
    parser.add_argument(
        '--ignore-namespace', '-n', action='append', default=[],
        help=('Namespace to be ignored. Can be given multiple '
              'times'))

    # Options to pass forward to unittest
    parser.add_argument(
        '--buffer', '-b',
        dest='unittest', action='append_const', const='-b',
        help=('The standard output and standard error streams are buffered '
              'during the test run. Output during a passing test is '
              'discarded. Output is echoed normally on test fail or error '
              'and is added to the failure messages.'))
    parser.add_argument(
        '--catch',
        dest='unittest', action='append_const', const='-c',
        help=('Control-C during the test run waits for the current test to '
              'end and then reports all the results so far. A second '
              'control-C raises the normal KeyboardInterrupt exception'))
    parser.add_argument(
        '--failfast', '-f',
        dest='unittest', action='append_const', const='-f',
        help=('Stop the test run on the first error or failure.'))
    try:
        args = parser.parse_args()
        sys.argv[1:] = args.unittest if args.unittest else []
        if args.verbose:
            sys.argv.append('-v')
    except SystemExit as e:
        os._exit(e.code)  # Exit without traceback on invalid arguments

    # Append to onto_path
    for paths in args.path:
        for path in paths.split(','):
            if path not in onto_path:
                onto_path.append(path)

    # Load ontology
    world = World(filename=args.database)
    if args.database != ':memory:' and args.iri not in world.ontologies:
        parser.error('The IRI argument should be one of the ontologies in '
                     'the database:\n  ' +
                     '\n  '.join(world.ontologies.keys()))

    onto = world.get_ontology(args.iri)
    onto.load(only_local=args.local,
              url_from_catalog=args.url_from_catalog,
              catalog_file=args.catalog_file)

    # Store settings TestEMMOConventions
    TestEMMOConventions.onto = onto
    TestEMMOConventions.check_imported = args.check_imported
    TestEMMOConventions.ignore_namespace = args.ignore_namespace

    # Configure tests
    verbosity = 2 if args.verbose else 1
    if args.configfile:
        import yaml
        with open(args.configfile, 'rt') as f:
            TestEMMOConventions.config.update(
                yaml.load(f, Loader=yaml.SafeLoader))

    # Run all subclasses of TestEMMOConventions as test suites
    status = 0
    for cls in TestEMMOConventions.__subclasses__():
        suite = unittest.TestLoader().loadTestsFromTestCase(cls)

        # Skip tests from command line
        for pattern in args.skip:
            for test in suite:
                name = test.id().split('.')[-1]
                if fnmatch.fnmatchcase(name, pattern):
                    setattr(test, 'setUp',
                            lambda: test.skipTest('skipped from command line'))

        runner = TextTestRunner(verbosity=verbosity)
        runner.resultclass.checkmode = True
        result = runner.run(suite)
        if result.failures:
            status = 1

    return status


if __name__ == '__main__':
    status = main()
    sys.exit(status)



"""Some generic utility functions.
"""
import os
import re
import datetime
import xml.etree.ElementTree as ET

from rdflib import Graph, URIRef

import owlready2


def asstring(expr, link='{name}', n=0, exclude_object=False):
    """Returns a string representation of `expr`, which may be an entity,
    restriction, or logical expression of these.  `link` is a format
    string for formatting references to entities or relations.  It may
    contain the keywords "name" and "url".
    `n` is the recursion depth and only intended for internal use.
    If `exclude_object` is true, the object will be excluded in restrictions.
    """
    def fmt(e):
        """Returns the formatted label of `e`."""
        name = str(e.label.first() if hasattr(e, 'label') and e.label else e)
        if re.match(r'^[a-z]+://', name):
            return link.format(name=name, url=name, lowerurl=name.lower())
        if hasattr(e, 'label') and e.label:
            name = e.label.first()
            url = name if re.match(r'^[a-z]+://', name) else '#' + name
            return link.format(name=name, url=url, lowerurl=url.lower())
        elif re.match(r'^[a-z]+://', str(e)):
            return link.format(name=e, url=e, lowerurl=e.lower())
        else:
            return str(e).replace('owl.', 'owl:')

    if isinstance(expr, str):
        # return link.format(name=expr)
        return fmt(expr)
    elif isinstance(expr, owlready2.Restriction):
        rlabel = owlready2.class_construct._restriction_type_2_label[expr.type]

        if isinstance(expr.property, (
                owlready2.ObjectPropertyClass,
                owlready2.DataPropertyClass)):
            s = fmt(expr.property)
        elif isinstance(expr.property, owlready2.Inverse):
            s = 'Inverse(%s)' % asstring(expr.property.property, link, n + 1)
        else:
            print('*** WARNING: unknown restriction property: %r' %
                  expr.property)
            s = fmt(expr.property)

        if not rlabel:
            pass
        elif expr.type in (owlready2.MIN, owlready2.MAX, owlready2.EXACTLY):
            s += ' %s %d' % (rlabel, expr.cardinality)
        elif expr.type in (owlready2.SOME, owlready2.ONLY,
                           owlready2.VALUE, owlready2.HAS_SELF):
            s += ' %s' % rlabel
        else:
            print('*** WARNING: unknown relation', expr, rlabel)
            s += ' %s' % rlabel

        if not exclude_object:
            if isinstance(expr.value, str):
                s += ' "%s"' % asstring(expr.value, link, n + 1)
            else:
                s += ' %s' % asstring(expr.value, link, n + 1)
        return s

    elif isinstance(expr, owlready2.Or):
        s = '%s' if n == 0 else '(%s)'
        return s % ' or '.join([asstring(c, link, n + 1)
                                for c in expr.Classes])
    elif isinstance(expr, owlready2.And):
        s = '%s' if n == 0 else '(%s)'
        return s % ' and '.join([asstring(c, link, n + 1)
                                 for c in expr.Classes])
    elif isinstance(expr, owlready2.Not):
        return 'not %s' % asstring(expr.Class, link, n + 1)
    elif isinstance(expr, owlready2.ThingClass):
        return fmt(expr)
    elif isinstance(expr, owlready2.PropertyClass):
        return fmt(expr)
    elif isinstance(expr, owlready2.Thing):  # instance (individual)
        return fmt(expr)
    elif isinstance(expr, owlready2.class_construct.Inverse):
        return fmt(expr)
    elif isinstance(expr, owlready2.disjoint.AllDisjoint):
        return fmt(expr)
    elif isinstance(expr, (bool, int, float)):
        return repr(expr)
    # Check for subclasses
    elif issubclass(expr, (bool, int, float, str)):
        return fmt(expr.__class__.__name__)
    elif issubclass(expr, datetime.date):
        return 'date'
    elif issubclass(expr, datetime.time):
        return 'datetime'
    elif issubclass(expr, datetime.datetime):
        return 'datetime'
    else:
        raise RuntimeError('Unknown expression: %r (type: %r)' % (
            expr, type(expr)))


def camelsplit(s):
    """Splits CamelCase string `s` before upper case letters (except
    if there is a sequence of upper case letters)."""
    if len(s) < 2:
        return s
    result = []
    prev_lower = False
    prev_isspace = True
    c = s[0]
    for next in s[1:]:
        if ((not prev_isspace and c.isupper() and next.islower()) or
                prev_lower and c.isupper()):
            result.append(' ')
        result.append(c)
        prev_lower = c.islower()
        prev_isspace = c.isspace()
        c = next
    result.append(next)
    return ''.join(result)


def read_catalog(path, catalog_file='catalog-v001.xml', recursive=False,
                 return_paths=False):
    """Reads a Protègè catalog file and returns a dict mapping IRIs to
    absolute paths.

    The `catalog_file` argument spesifies the catalog file name and is
    used if `path` is used when `recursive` is true or when `path` is a
    directory.

    If `recursive` is true, catalog files in sub-folders are also read.

    If `return_paths` is true, a set of directory paths to source
    files is returned in addition to the default dict.
    """
    iris = {}
    dirs = set()
    if os.path.isdir(path):
        dirname = os.path.abspath(path)
        filepath = os.path.join(dirname, catalog_file)
    else:
        catalog_file = os.path.basename(path)
        filepath = os.path.abspath(path)
        dirname = os.path.dirname(filepath)

    def gettag(e):
        return e.tag.rsplit('}', 1)[-1]

    def load_catalog(filepath):
        dirname = os.path.normpath(os.path.dirname(filepath))
        dirs.add(dirname)
        xml = ET.parse(filepath)
        root = xml.getroot()
        if gettag(root) != 'catalog':
            raise ValueError('expected root tag of catalog file %r to be '
                             '"catalog"', filepath)
        for child in root:
            if gettag(child) == 'uri':
                load_uri(child, dirname)
            elif gettag(child) == 'group':
                for uri in child:
                    load_uri(uri, dirname)

    def load_uri(uri, dirname):
        assert gettag(uri) == 'uri'
        s = uri.attrib['uri']
        if s.startswith('http://') or s.startswith('https://'):
            iris.setdefault(uri.attrib['name'], s)
        else:
            filepath = os.path.join(dirname, uri.attrib['uri'])
            iris.setdefault(uri.attrib['name'], filepath)
            dir = os.path.normpath(os.path.dirname(filepath))
            if recursive and dir not in dirs:
                catalog = os.path.join(dir, catalog_file)
                load_catalog(catalog)

    load_catalog(filepath)
    if return_paths:
        return iris, dirs
    else:
        return iris


def convert_imported(input, output, input_format=None, output_format='xml',
                     catalog_file='catalog-v001.xml'):
    """Convert imported ontologies.

    Store the output in a directory structure matching the source
    files.  This require catalog file(s) to be present.

    Args:
        input: input ontology file name
        output: output ontology file path.  The directory part of `output`
            will be the root of the generated directory structure
        input_format: input format.  The default is to infer from `input`
        output_format: output format.  The default is to infer from `output`
        catalog_file: name of catalog file, that maps ontology IRIs to
            local file names
    """
    inroot = os.path.dirname(os.path.abspath(input))
    outroot = os.path.dirname(os.path.abspath(output))
    d, dirs = read_catalog(inroot, catalog_file=catalog_file, recursive=True,
                           return_paths=True)
    # Create output dirs and copy catalog files
    outext = os.path.splitext(output)[1]
    for indir in dirs:
        outdir = os.path.normpath(
            os.path.join(outroot, os.path.relpath(indir, inroot)))
        if not os.path.exists(outdir):
            os.makedirs(outdir)
        with open(os.path.join(indir, catalog_file), mode='rt') as f:
            s = f.read()
        for path in d.values():
            newpath = os.path.splitext(path)[0] + outext
            s = s.replace(os.path.basename(path), os.path.basename(newpath))
        with open(os.path.join(outdir, catalog_file), mode='wt') as f:
            f.write(s)

    outpaths = set()
    def recur(graph, outext):
        for imported in graph.objects(predicate=URIRef(
                'http://www.w3.org/2002/07/owl#imports')):
            inpath = d[str(imported)]
            if inpath.startswith('http://'):
                outpath = inpath.split('/')[-1]
            elif inpath.startswith('https://'):
                outpath = inpath.split('/')[-1]
            else:
                outpath=inpath
            outpath = os.path.splitext(os.path.normpath(
                os.path.join(outroot, os.path.relpath(
                    outpath, inroot))))[0] + outext
            if outpath not in outpaths:
                outpaths.add(outpath)
                g = Graph()
                g.parse(inpath, format=input_format)
                g.serialize(destination=outpath, format=output_format)
                recur(g, outext)

    # Write output files
    g = Graph()
    g.parse(input, format=input_format)
    g.serialize(destination=output, format=output_format)
    recur(g, outext)


def squash_imported(input, output, input_format=None, output_format='xml',
                     catalog_file='catalog-v001.xml'):
    """Convert imported ontologies and squash them into a single file.

    If a catalog file exists in the same directory as the input file it will
    be used to load possible imported ontologies.
    """
    inroot = os.path.dirname(os.path.abspath(input))
    if catalog_file and os.path.exists(os.path.join(inroot, catalog_file)):
        d = read_catalog(inroot, catalog_file=catalog_file, recursive=True)
    else:
        d = {}

    imported = set()
    def recur(g):
        for s, p, o in g.triples(
                (None, URIRef('http://www.w3.org/2002/07/owl#imports'), None)):
            g.remove((s, p, o))
            iri = d.get(str(o), str(o))
            if iri not in imported:
                imported.add(iri)
                g2 = Graph()
                g2.parse(iri, format=input_format)
                recur(g2)
                for t in g2.triples((None, None, None)):
                    graph.add(t)

    graph = Graph()
    graph.parse(input, format=input_format)
    recur(graph)
    graph.serialize(destination=output, format=output_format)


def infer_version(iri, version_iri):
    """Infer version from IRI and versionIRI."""
    if str(version_iri[:len(iri)]) == str(iri):
        version = version_iri[len(iri):].lstrip('/')
    else:
        j = 0
        v = []
        for i in range(len(iri)):
            while i + j < len(version_iri) and iri[i] != version_iri[i + j]:
                v.append(version_iri[i + j])
                j += 1
        version = ''.join(v).lstrip('/').rstrip('/#')

    if '/' in version:
        raise ValueError('version IRI %r is not consistent with base IRI '
                         '%r' % (version_iri, iri))
    return version



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import sys
from glob import glob
import unittest

thisdir = os.path.dirname(__file__)


# Wrap tests into a unittest TestCase
# This way, we can run the individual tests interactively
class ScriptTestCase(unittest.TestCase):
    def __init__(self, methodname='run_tests', filename=None):
        unittest.TestCase.__init__(self, methodname)
        self.filename = filename

    def run_tests(self):
        env = globals().copy()
        env.update(__file__=self.filename)
        with open(self.filename) as fd:
            exec(compile(fd.read(), self.filename, 'exec'), env)

    def id(self):
        return self.filename

    def __str__(self):
        return self.filename.split('tests/')[-1]

    def __repr__(self):
        return "ScriptTestCase(filename='%s')" % self.filename


def test(verbosity=1, stream=sys.stdout):
    tests = [test for test in glob(os.path.join(thisdir, '*.py'))
             if not test.endswith('__.py') and
             not test.endswith(os.path.basename(__file__))]
    ts = unittest.TestSuite()
    for test in tests:
        ts.addTest(ScriptTestCase(filename=os.path.abspath(test)))
    with open(os.devnull, 'w') as devnull:
        if not verbosity:
            stream = devnull
        ttr = unittest.TextTestRunner(verbosity=verbosity, stream=stream)

        # Redirect stdout and stderr to devnull
        # stderr is redicred at file-descriptor level to get rid of
        # C-level output
        dest_fd = devnull.fileno()
        stderr_fd = sys.stderr.fileno()
        # copy stderr_fd before it is overwritten
        # NOTE: `copied` is inheritable on Windows when duplicating a
        # standard stream
        with os.fdopen(os.dup(stderr_fd), 'wb') as copied:
            sys.stdout.flush()
            sys.stderr.flush()
            try:
                sys.stdout = devnull
                os.dup2(dest_fd, stderr_fd)  # $ exec >&dest
                results = ttr.run(ts)
            finally:
                sys.stdout.flush()
                sys.stderr.flush()
                sys.stdout = sys.__stdout__
                os.dup2(copied.fileno(), stderr_fd)  # $ exec >&copied
    return results


if __name__ == "__main__":
    # unittest.main()
    results = test()
    if results.errors or results.failures:
        sys.exit(1)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Step 3 - load atom structure and represent it using our metadata framework
--------------------------------------------------------------------------
In this step we uses the Atomistic Simulation Environment (ASE) to load
a atomistic Al-Fe4Al13 interface structure from a cif file and
represents it using the metadata defined in step 2.
"""
import ase
import ase.io

import dlite

from step2_define_metadata import DLiteAtoms


# Example for importing structure from the MaterialsProject.
# See https://materialsproject.org/docs/api for instructions for getting
# personal access.
# Not used in the current demo
#
#     from pymatgen.ext.matproj import MPRester
#     with MPRester('USER_API_KEY') as m:
#         structure = m.get_structure_by_material_id('mp-2018')
#


# Load atom structure from cif file
at = ase.io.read('../vertical/Al-Fe4Al13.cif')

# convert atom structure to a DLiteAtoms object
atoms = dlite.objectfactory(at, cls=DLiteAtoms, instanceid='atoms_Al-Fe4Al13')


# Create a new collection for data instances
coll = dlite.Collection('usercase_appdata')
coll.add('Atoms', atoms.dlite_meta)
coll.add('atoms', atoms.dlite_inst)
coll.save('json', 'usercase_appdata.json', 'mode=w')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""An example script that uses EMMO to describe a vertical use case on
welding aluminium to steel and how the thin layer of intermetallic
that are formed at the interface is influencing the overall
properties.  Based on TEM observations using scanning precision
electron diffraction (SPED) the following (simplified) sequence of
intermetallic phases could be established:

    Al | alpha-AlFeSi | Fe4Al13 | Fe2Al5 | Fe

which is consistent with phase stability when the Fe-concentration is
increasing when going from left to right.

In this case study three scales are considered:

  - Macroscopic scale: predicts the overall mechanical behaviour of
    the welded structure during deformation.

  - Microscopic scale: a local crystal plasticity model of a small
    part of the interface.  The constitutive equations were based on
    the results from DFT.  The results from this model was used to
    calibrate decohesion elements for the macroscopic scale.

  - Electronic scale: elastic properties of the individual phases as
    well work of decohesion within and at the interfaces between the
    phases were calculated with DFT [1].  The calculation of work of
    decohesion was performed as a series of rigid steps, providing
    stress-strain relations in both tensile and shear.

References
----------
[1] Khalid et al. Proc. Manufact. 15 (2018) 1407

"""
from emmo import World


# Load EMMO
world = World(filename='demo.sqlite3')
emmo = world.get_ontology('http://emmo.info/emmo/1.0.0-alpha2')
emmo.load()
# emmo.sync_reasoner()

# Create a new ontology with out extensions that imports EMMO
onto = world.get_ontology('http://www.emmc.info/emmc-csa/demo#')
onto.imported_ontologies.append(emmo)


# Add new classes and object/data properties needed by the use case
with onto:

    #
    # Relations
    # =========
    class hasType(emmo.hasConvention):
        """Associates a type (string, number...) to a property."""
        pass

    class isTypeOf(emmo.hasConvention):
        """Associates a property to a type (string, number...)."""
        inverse_property = hasType

    #
    # Units
    # =====

    # TODO: remove
    class SquareLengthDimension(emmo.PhysicalDimension):
        is_a = [emmo.hasSymbolData.value('T0 L2 M0 I0 Θ0 N0 J0')]

    # TODO: remove
    class SquareMetre(emmo.SICoherentDerivedUnit):
        emmo.altLabel = ['m²']
        is_a = [emmo.hasPhysicalDimension.only(SquareLengthDimension)]

    #
    # Properties
    # ==========

    # TODO: update instead of redefine Position
    class Position(emmo.Length):
        """Spatial position of an physical entity."""
        is_a = [emmo.hasReferenceUnit.only(emmo.hasPhysicalDimension.only(
                 emmo.LengthDimension)),
                hasType.exactly(3, emmo.Real)]

    # TODO: remove
    class Area(emmo.ISQDerivedQuantity):
        """Extent of a surface."""
        is_a = [
            emmo.hasReferenceUnit.only(emmo.hasPhysicalDimension.only(
                SquareLengthDimension)),
            hasType.exactly(1, emmo.Real),
        ]

    emmo.Pressure.is_a.append(hasType.exactly(1, emmo.Real))

    # TODO: update when we have dimensionality
    class StiffnessTensor(emmo.Pressure):
        r"""The stiffness tensor $c_{ijkl}$ is a property of a continuous
        elastic material that relates stresses to strains (Hooks's
        law) according to

            $\sigma_{ij} = c_{ijkl} \epsilon_{kl}$

        Due to symmetry and using the Voight notation, the stiffness
        tensor can be represented as a symmetric 6x6 matrix

            / c_1111  c_1122  c_1133  c_1123  c_1131  c_1112 \
            | c_2211  c_2222  c_2233  c_2223  c_2231  c_2212 |
            | c_3311  c_3322  c_3333  c_3323  c_3331  c_3312 |
            | c_2311  c_2322  c_2333  c_2323  c_2331  c_2312 |
            | c_3111  c_3122  c_3133  c_3123  c_3131  c_3112 |
            \ c_1211  c_1222  c_1233  c_1223  c_1231  c_1212 /

        """
        is_a = [hasType.exactly(36, emmo.Real)]

    # class Spacegroup(emmo.DescriptiveProperty):
    #     """A spacegroup is the symmetry group off all symmetry operations
    #     that apply to a crystal structure.
    #
    #     It is identifies by its Hermann-Mauguin symbol or space group
    #     number (and setting) in the International tables of
    #     Crystallography."""
    #     is_a = [hasType.exactly(1, emmo.String)]
    #     pass

    # class Plasticity(emmo.PhysicalQuantity):
    #     """Describes Yield stress and material hardening."""
    #     is_a = [hasUnit.exactly(1, Pascal),
    #             hasType.min(2, emmo.Real)]

    ''' Will be included when dimensionality is inplace in EMMO'''

    # class TractionSeparation(Pressure):
    #     """The force required to separate two materials a certain distance
    #     per interface area.  Hence, traction_separation is a curve, that
    #     numerically can be represented as a series of (force,
    #     separation_distance) pairs."""
    #     is_a = [hasUnit.exactly(1, Pascal),
    #             hasType.min(4, emmo.Real)]

    # class LoadCurve(Pressure):
    #     """A measure for the displacement of a material as function of the
    #     appliced force."""
    #     is_a = [hasUnit.exactly(1, Pascal),
    #             hasType.min(4, emmo.Real)]

    # Crystallography-related classes
    # TODO: import crystallography ontology instead
    # -------------------------------
    class LatticeVector(emmo.Length):
        """A vector that participitates defining the unit cell."""
        is_a = [hasType.exactly(3, emmo.Real)]

    # FIXME - CrystalUnitCell is not a matter, but a model or a symbolic
    #         Just use crystalography
    class CrystalUnitCell(emmo.Material):
        """A volume defined by the 3 unit cell vectors.  It contains the atoms
        constituting the unit cell of a crystal."""
        is_a = [emmo.hasSpatialDirectPart.some(emmo.BondedAtom),
                emmo.hasProperty.exactly(3, LatticeVector),
                emmo.hasProperty.exactly(1, StiffnessTensor)]

    class InterfaceModel(CrystalUnitCell):
        is_a = [emmo.hasProperty.some(Area)]

    class Crystal(emmo.Solid):
        """A periodic crystal structure."""
        is_a = [emmo.hasSpatialDirectPart.only(CrystalUnitCell)]

    # Add some properties to our atoms
    emmo.Atom.is_a.append(emmo.hasProperty.exactly(1, Position))

    # Continuum
    # ---------
    class Boundary(emmo.Continuum):
        """A boundary is a 4D region of spacetime shared by two material
        entities."""
        equivalent_to = [emmo.hasSpatialDirectPart.exactly(2, emmo.Continuum)]
        is_a = [emmo.hasProperty.exactly(1, Area)]

    class Phase(emmo.Continuum):
        """A phase is a continuum in which properties are homogeneous and can
        have different state of matter."""
        is_a = [emmo.hasProperty.exactly(1, StiffnessTensor)]

    class RVE(emmo.Continuum):
        """Representative volume element.  The minimum volume that is
        representative for the system in question."""
        is_a = [emmo.hasSpatialDirectPart.only(Phase | Boundary)]

    class WeldedComponent(emmo.EngineeredMaterial):
        """A welded component consisting of two materials welded together
        using a third welding material.  Hence it has spatial direct
        parts 3 materials and two boundaries."""
        is_a = [
            emmo.hasSpatialDirectPart.exactly(3, emmo.Material),
            emmo.hasSpatialDirectPart.exactly(2, Boundary),
            # emmo.hasProperty.exactly(1, LoadCurve),
        ]


# Sync attributes to make sure that all classes get a `label` and to
# include the docstrings in the comments
onto.sync_attributes(name_policy='uuid', name_prefix='DEMO_')


# Run the reasoner
# onto.sync_reasoner()

# set version of ontology
onto.set_version("0.9")

# Save our new EMMO-based ontology to demo.owl
onto.save('demo.owl', overwrite=True)

# ...and to the sqlite3 database.
world.save()



#!/usr/bin/env python3
"""
Python reference API for the Europeean Materials & Modelling Ontology (EMMO).
"""
import os
import re
import setuptools
from glob import glob


rootdir = os.path.dirname(__file__)


def rglob(patt):
    """Recursive glob function that only returns ordinary files."""
    return [f for f in glob(patt, recursive=True) if os.path.isfile(f)]


def fglob(patt):
    """Glob function that only returns ordinary files."""
    return [f for f in glob(patt) if os.path.isfile(f) and not f.endswith('~')]


# Read long description from README.md file replacing references to local
# files to github urls
baseurl = 'https://raw.githubusercontent.com/emmo-repo/EMMO-python/master/'
with open(os.path.join(rootdir, 'README.md'), 'rt') as f:
    long_description = re.sub(
        r'(\[[^]]+\])\(([^:)]+)\)', rf'\1(%s\2)' % baseurl, f.read())

# Read requirements from requirements.txt file
with open(os.path.join(rootdir, 'requirements.txt'), 'rt') as f:
    requirements = f.read().split()

# Retrieve emmo-package version
with open(os.path.join(rootdir, 'emmo/__init__.py')) as handle:
    for line in handle:
        match = re.match(r"__version__ = '(.*)'", line)
        if match is not None:
            VERSION = match.group(1)
            break
    else:
        raise RuntimeError(f'Could not determine package version from {handle.name} !')


setuptools.setup(
    name='EMMO',
    version=VERSION,
    author='Jesper Friis, Francesca Lønstad Bleken, Bjørn Tore Løvfall',
    author_email='jesper.friis@sintef.no',
    description=('Python reference API for the Europeean Materials & '
                 'Modelling Ontology'),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/emmo-repo/EMMO-python',
    license='BSD',
    python_requires='>=3.6.0',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    install_requires=requirements,
    packages=['emmo'],
    scripts=['tools/ontodoc', 'tools/ontograph', 'tools/emmocheck',
             'tools/ontoconvert', 'tools/ontoversion'],
    package_data={'emmo': ['tests/*.py']},
    data_files=[
        ('share/EMMO-python', ['README.md', 'LICENSE.txt']),
        (
            'share/EMMO-python/examples/emmodoc',
            glob('examples/emmodoc/*.md') +
            glob('examples/emmodoc/*.yaml') +
            glob('examples/emmodoc/pandoc-*'),
        ),
        (
            'share/EMMO-python/examples/emmodoc/figs',
            fglob('examples/emmodoc/figs/*'),
        ),
        # ('share/EMMO-python/examples', rglob('examples/**')),
        ('share/EMMO-python/demo', rglob('demo/**')),
    ],
)
